from json import loads


class BaseModel(object):

    @classmethod
    def from_json(cls, json):
        return cls.from_dict(loads(json))

    @classmethod
    def from_dict(cls, dict_):
        instance = cls()
        for key in dict_:
            setattr(instance, key, dict_[key])
        return instance

    @classmethod
    def from_list(cls, list_):
        model_list = None
        if list_ is not None:
            model_list = []
            for item in list_:
                model_list.append(cls.from_dict(item))
        return model_list

    @classmethod
    def from_json_list(cls, json_list):
        return cls.from_list(loads(json_list))

    @staticmethod
    def _model_list(class_):
        """
        :param class_: The class that elements should be an instance of.
        :return: A decorator that ensures the assigning value is a list of `class_` instances.
        """
        assert issubclass(class_, BaseModel)

        def decorator(f):
            def wrapper(self, list_):
                if isinstance(list_, list):
                    model_list = []
                    for item in list_:
                        if isinstance(item, class_):
                            model_list.append(item)
                        elif isinstance(item, dict):
                            model_list.append(class_.from_dict(item))
                        else:
                            raise HawkIOException(u'Invalid element type.')
                    f(self, model_list)
                else:
                    f(self, [])
            return wrapper
        return decorator

    @staticmethod
    def _model(class_):
        """
        :param class_: The class that the value should be an instance of.
        :return: A decorator that ensures the assigning value is of `class_` instances.
        """
        assert issubclass(class_, BaseModel)

        def decorator(f):
            def wrapper(self, model):
                if isinstance(model, class_):
                    f(self, model)
                elif isinstance(model, dict):
                    f(self, class_.from_dict(model))
                elif model is None:
                    f(self, None)
                else:
                    raise HawkIOException(u'Invalid value type.')
            return wrapper
        return decorator

    # def as_payload(self, filter_=None):
        # return payload_filter(self.__dict__, filter_)


class Filter(BaseModel):

    def __init__(
            self,
            operators=None,
            control=None,
            name=None,
            readable_name=None
    ):
        self.operators = operators
        self.control = control
        self.name = name
        self.readable_name = readable_name


class Filters(BaseModel):

    def __init__(
            self,
            filters=None,
            sort=None,
            wildcard_fields=None
    ):
        self._filters = None
        self.filters = filters
        self.sort = sort
        self.wildcard_fields = wildcard_fields

    @property
    def filters(self):
        return self._filters

    @filters.setter
    @BaseModel._model_list(Filter)
    def filters(self, filters):
        self._filters = filters


class FilterSort(BaseModel):

    def __init__(
            self,
            name=None,
            order=None
    ):
        self.name = name
        self.order = order


class FilterPagination(BaseModel):

    def __init__(
            self,
            total=None,
            offset=None,
            limit=None,
            sort=None
    ):
        self._sort = None
        self.total = total
        self.offset = offset
        self.limit = limit
        self.sort = sort

    @property
    def sort(self):
        return self._sort

    @sort.setter
    @BaseModel._model_list(FilterSort)
    def sort(self, sorts):
        self._sort = sorts


# utility #wait/sleep #add to enrichment list.



# for interacting with eyecon api, like logging in, getting events, agent details.

import requests
# HAWK API
import hawkAPI.lib.core.SmtpSend as SmtpSend
from hawkAPI.lib.core.ThreadPool import *

from hawkAPI.lib.core.hawkcore import hawkcore
from hawkAPI.lib.core.hawkapi import hawkapi
from hawkAPI.lib.core.hawklib import hawklib

from errors import *


class eyeconAPI(object):
    def __init__(self, config):
        self.config = config

    def login(self):
        """ Login into HAWK API."""

        # authenticate and start our handler

        try:
                self.hawk = hawkcore(self.config['api_server'])
                self.hawk.logfile("/var/log/hawk/hawk-health.log")
        except Exception as e:
                errstr = 'Failure: Unable to create hawkcore object using API Server: %s' % (
                    self.config['api_server'])
                errors = getattr(e, 'errors', None)
                logger.exception(errstr)
                raise HawkCoreError(
                    errstr, errors, api_server=self.config['api_server'])
        try:
                if 'insecure' in self.config.keys() and self.config['insecure'].lower() == 'true':
                        self.hawk.SSLVerify(False)
                if 'timeout' in self.config.keys():
                        self.hawk.setTimeout(self.config['timeout'])
                if 'retry' in self.config.keys():
                        self.hawk.setRetry(self.config['retry'])
                if 'verbose' in self.config.keys() and self.config['verbose']:
                        self.hawk.debug(1)
        except Exception as e:
                errstr = 'Failure: Unable to set attributes.'
                errors = getattr(e, 'errors', None)
                logger.exception(errstr)
                raise HawkCoreError(errstr, errors)

        try:
                x = self.hawk.login(
                    self.config['api_username'], self.config['api_password'])
                if x == None:
                        raise Exception("authentication failed")
                logger.info("Authenticated with remote API successfully (%s @ %s) ..." %
                            (self.config['api_username'], self.config['api_server']))
                self.config['user_info'] = x
        except Exception as e:
                errstr = 'Failure: User %s Failed to log into API %s (%s)' % (
                    self.config['api_username'], self.config['api_server'], e)
                errors = getattr(e, 'errors', None)
                logger.exception(errstr)
                raise HawkLoginError(
                    errstr, errors, user=self.config['api_username'])

    def query_events(self, start_time=None, end_time=None, **kwargs):
        """ Query for events.
        Args:
            start_time (DATETIME): Start time for query
            end_time (DATETIME): End time for query
            **kwargs: Arbitrary keyword arguments for querying events.
        Returns:
            json: results of event query for success. False, if unsuccessful.
        """
        pass

    def query_agents(self, uuid=None, hostname=None, address=None, group=None):
        # https://portal.hawk.io:8080/API/1.1/search/agents
        """Query HAWK Agents
        List all agents, or query for specific agents.

        Example:
            self.query_agents()

        Args:
            uuid (str): Agent UUID.
            hostname (str): Agent's Hostname.
            address (str): Agent's IPv4 IP Address.
            group (str): Group Agent is located in.

        Returns:
            (list) List of dictionaries of agents.
        """

        # Query local DB first, If doesn't exist then query API.

        # Always Query all agents, and do the filter locally.

        # Save all the agents in our local database.

        # Return
'''[{
    'address': '54.202.126.194',
    'agent_id': 91,
    'agent_uuid': 'vol046af7ba7853167c1',
    'agent_version': 'v1.6.0-94504',
    'applications': [ {'install_date': '20190906', 'location': '', 'name': 'Amazon SSM Agent', 'publisher': 'Amazon Web Services', 'version': '2.3.634.0'}],
    'approved_user_id': null,
    'date_added': '2019-10-30 12:58:27',
    'date_approved': null,
    'date_updated': '2019-11-26 21:25:46',
    'group_name': 'hawk',
    'hostname': 'EC2AMAZ-S5GK1EM',
    'os': 'Microsoft Windows Server 2019 Datacenter x64 (6.3 build 17763)',
    'pending_updates': false,
    'settings': [],
    'status': true,
    'system_info': {cpu: {}, 'disk': {}, 'hostname': '', 'memory': {}, 'network': {}, 'processes': {}, 'services': [{}], 'tasks': []},
    'updates': [],
}]'''
        pass

    def agent_details(self, uuid=None):
        """Query HAWK agent for details.

        Args:
            uuid (str): Agent UUID.

        Returns:
            (dict) Agent details like network connections.
        """

        # Query API for agent details

        # Save results in DB, with cache enter datetime.

        # return lookup point (This is to simplefy the process and to limit memory usage)

        pass

    def get_playbooks(self):
        """ Get and Parse Playbooks

            Returns:
                list of playbooks
        """

        # Query playbooks
        #




from sqlalchemy import create_engine


from hawkSoar.common.logger import get_logger


class DBInterface(object):
    """
    This is the Sqlite interface base class handling loading config and setup connections
    """

    def __init__(self, config=None):
        self.engines = {
            'sqlite': 'sqlite:///{DB}',
            'mysql': 'mysql://{USER}:{PASSWORD}@{HOST}/{DB}'
        }

        self.logger = get_logger()

        self.config = config
        self.client = self._connect()
        self._authenticate()
        self._setup_database_mapping()

    def _connect(self):
        db_engine = None
        if self.config['DB_TYPE'] == 'sqlite':
            engine_url = self.engines[self.config['DB_TYPE']].format(
                DB=self.config['DB_NAME'])
            db_engine = create_engine(engine_url)

        elif self.config['DB_TYPE'] == 'mysql':
            engine_url = self.engines[self.config['DB_TYPE']].format(USER=self.config['USERNAME'],
                                                                     PASSWORD=self.config['PASSWORD'],
                                                                     HOST=self.config['HOST'],
                                                                     DB=self.config['DB_NAME'])
            db_engine = create_engine(engine_url)

        else:
            self.logger.error('DB_TYPE %s is not found in DB_ENGINE: %s' % (
                self.config['DB_TYPE'], db_engine))

        self.logger.info('Database interface opened for: %s %s' %
                         (self.config['DB_TYPE'], db_engine))
        return db_engine

    def shutdown(self):
        # close doesnt exist
        if self.client:
            self.client.dispose()

    def _setup_database_mapping(self):
        pass

    def _authenticate(self):
        pass





import os
import sys
import json
import requests
import time
from urllib.parse import urlparse

import xml
import radix
import csv
import zipfile
import pickle

import shutil

from hawkSoar.common.logger import get_logger
from hawkSoar.common.process import ExceptionSafeProcess
from hawkSoar.plugins.base import BasePlugin
from hawkSoar.lib.query_parser import QueryInterpreter


from pymisp import PyMISP


class InvestigatePlugin(BasePlugin):

    ID = 'misp-url'
    NAME = 'MISP URL'
    DESCRIPTION = "The MISP threat sharing platform is a free and open source software helping information sharing of threat intelligence including cyber security indicators.\n A threat intelligence platform for gathering, sharing, storing and correlating Indicators of Compromise of targeted attacks, threat intelligence, financial fraud information, vulnerability information or even counter-terrorism information."
    TYPE = 'investigate'
    DEPENDENCIES = ['PyMISP']
    VERSION = '0.0.1'
    MODULE = "playbook.investigate.URLReputation"
    ENABLED = True
    Exceptions = []
    playbookCache = {}
    activeRequests = []
    activeURL = {}

    def __init__(self, plugin_administrator, config=None):
        super().__init__(plugin_administrator, config=config, plugin_path=__file__)
        self.config = config
        self.logger = get_logger()
        self.playbookCache = {}

    def process_task(self, task):
        self.logger.debug("MISP URL TASK RECEIVED: %s" % task['cmd'])
        self.logger.debug(json.dumps(task))
        if task['cmd'] == 'addPlaybookStep':
            self._add_playbook_step(task['data'])
            # start up new worker when new task arrives.

            if not task['data']['id'] in self.playbookCache:
                self.playbookCache[task['data']['id']] = {}

            self.logger.debug("MATCH NEW STEP TASK")
            playbookId = task['data']['id']
            stepId = task['data']['step']['id']
            self.logger.debug(task['data']['step'])
            self.playbookCache[playbookId][stepId] = {
                'config':  task['data']['config'],
                'step':  task['data']['step']
            }
            stepId = task['data']['step']

            step = self.playbookCache[playbookId][stepId]

            # print(self.get_step_by_id(task['data']['playbookId'], task['data']['step']))

            task['data']['data']['artifacts'].append(
                {"plugin_id": self.ID, "name": "version", "module": self.MODULE, "type": "app.version",  "value": self.VERSION})

            result = None

            stepObj = self.get_step_by_id(
                task['data']['playbookId'], task['data']['step'])

            task['data']['data']['return_value'] = True

            task['route'] = 'publish'

            task['data']['step'] = step['step']['next']

            #self.logger.debug("OUTGOING ENRICHMENT TASK")
            # self.logger.debug(json.dumps(task))

            self.outgoing_queue.put(task)

    def add_exception(self, msg):
        self.Exceptions.append(msg)

    def check_exceptions(self):
        # if exceptiosn found, then log out
        return self.Exceptions

    def shutdown(self):
        # stop all on-going workers here
        self.logger.debug('MISP-URL plugin is now shutting down.')
        # super().shutdown()

        self.logger.debug('MISP-URL shutdown complete.')



from .base import BaseModel


class IngestObject(BaseModel):
    """ Data model for Investigate Objects.

        Args:
            uid (uuidv4): Unique uuidv4 of this obj instance, Assigned by SOAR API
            groups (dict): Group informations used for permissions
            status (str): local current status, can be one of the following new, pending, processing, done, timeout
            ttl (int): Seconds this object is allowed to live for. If ttl is exceeded object.status should be set to timeout and returned
            plugin_action (list): List of actions that should be performed and by what plugin(s) i.e.
                                {
                                    'plugin_id': plugin_id, #Plugin ID as it's registered in SOAR API
                                    'plugin_name': plugin_name,
                                    'plugin_config': plugin_config,
                                    'plugin_action': plugin_action
                                }
            processed_artifacts (dict): Results of plugin after processing request
    """

    def __init__(
            self,
            uid=None,
            groups=None,
            status=None,
            ttl=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.groups = groups
        self.status = status
        self.ttl = ttl
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class InvestigateObject(BaseModel):
    """ Data model for Investigate Objects.

        Args:
            uid (uuidv4): Unique uuidv4 of this obj instance, Assigned by SOAR API
            groups (dict): Group informations used for permissions
            status (str): local current status, can be one of the following new, pending, processing, done, timeout
            ttl (int): Seconds this object is allowed to live for. If ttl is exceeded object.status should be set to timeout and returned
            plugin_action (list): List of actions that should be performed and by what plugin(s) i.e.
                                {
                                    'plugin_id': plugin_id, #Plugin ID as it's registered in SOAR API
                                    'plugin_name': plugin_name,
                                    'plugin_config': plugin_config,
                                    'plugin_action': plugin_action
                                }
            artifacts (list): List of dictionaries containing the information plugins can use to process the request. TODO:Still in progress i.e.
                                {
                                    event_data : {
                                        resource_address : '127.0.0.1',
                                        key : value
                                    },
                                    plugin_data : {
                                        key : value,
                                        key : value
                                    },
                                    other : {
                                        key:value,
                                        key:value
                                    }
                                }
            processed_artifacts (dict): Results of plugin after processing request
    """

    def __init__(
            self,
            uid=None,
            groups=None,
            status=None,
            ttl=None,
            plugin_action=[],
            artifacts=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.ttl = ttl
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.artifacts = artifacts
        self.processed_artifacts = processed_artifacts


class ContainObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class RecoverObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class NotifyObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class RecordObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class UtilityObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class DecisionObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts


class IncludeObject(BaseModel):
    def __init__(
            self,
            uid=None,
            plugin_id=None,
            soar_id=None,
            groups=None,
            status=None,
            plugin_action=[],
            processed_artifacts={}
    ):
        self.uid = uid
        self.plugin_id = plugin_id
        self.soar_id = soar_id
        self.groups = groups
        self.status = status
        self.plugin_action = plugin_action
        self.processed_artifacts = processed_artifacts





import os
import sys
import json

from hawkSoar.common.logger import get_logger
from hawkSoar.common.process import ExceptionSafeProcess
from hawkSoar.plugins.base import BasePlugin
from hawkSoar.lib.query_parser import QueryInterpreter


class DecisionPlugin(BasePlugin):

    ID = 'decision'
    NAME = 'Decision'
    DESCRIPTION = 'Decision data using provided result'
    TYPE = 'decision'
    DEPENDENCIES = []
    VERSION = '1.0.0'
    MODULE = 'playbook.decision.Decision'
    ENABLED = True
    Exceptions = []
    playbookCache = {}
    DecisionToStep = {}

    def __init__(self, plugin_administrator, config=None):
        super().__init__(plugin_administrator, config=config, plugin_path=__file__)
        self.config = config
        self.logger = get_logger()
        self.playbookCache = {}
        self.DecisionToStep = {}

    def process_task(self, task):
        if task['cmd'] == 'addPlaybookStep':
            # self.logger.debug(json.dumps(task))

            self._add_playbook_step(task['data'])
            # start up new worker when new task arrives.

            if not task['data']['id'] in self.playbookCache:
                self.playbookCache[task['data']['id']] = {}

            self.playbookCache[task['data']['id']][task['data']['step']['id']] = {
                'config':  task['data']['step']
            }

            # for each next step...
            for next_step in task['data']['step']['next']:
                nx = self.get_step_by_id(task['data']['id'], next_step)
                dec = nx['decision']
                step = task['data']['id'] + ':' + nx['id']
                if not step in self.DecisionToStep:
                    self.DecisionToStep[step] = {"yes": [], "no": []}

                self.DecisionToStep[step][dec].append(nx['id'])

        else:
            # do the work
            # use return_value to make the decision of which step(s) go next

            task['route'] = 'publish'

            # this part depends
            step = self.get_step_by_id(
                task['data']['playbookId'], task['data']['step'])
            for next_step in step['next']:
                step = task['data']['playbookId'] + ':' + next_step

                if 'return_value' in task['data']['data'] and task['data']['data']['return_value'] == True:
                    task['data']['step'] = self.DecisionToStep[step]["yes"]
                    if len(task['data']['step']) > 0:
                        #self.logger.debug("DECISION OUTBOUND YES TASK")
                        # self.logger.debug(task)
                        self.outgoing_queue.put(task)
                    # else this goes to /dev/null :)
                elif 'return_value' in task['data']['data'] and task['data']['data']['return_value'] == False:
                    task['data']['step'] = self.DecisionToStep[step]["no"]
                    if len(task['data']['step']) > 0:
                        #self.logger.debug("DECISION OUTBOUND NO TASK")
                        # self.logger.debug(task)
                        self.outgoing_queue.put(task)
                    # else this goes to /dev/null :)
                else:
                    raise Exception(
                        "Decision task does not havea areturn value, task: %s" % json.dumps(task))

    def add_exception(self, msg):
        self.Exceptions.append(msg)

    def check_exceptions(self):
        # if exceptiosn found, then log out
        return self.Exceptions

    def shutdown(self):
        # stop all on-going workers here
        self.logger.debug('Decision plugin is now shutting down.')
        # super().shutdown()
        self.logger.debug('Decision shutdown complete.')




__version__ = '0.2'
__author__ = 'Jason Wheeler'


def to_unicode(obj, encoding='utf-8', errors='replace'):
    ''' convert a 'str' to 'unicode' '''
    if isinstance(obj, basestring):
        if not isinstance(obj, unicode):
            obj = unicode(obj, encoding, errors)
    return obj


def to_utf8(obj, errors='replace'):
    '''convert 'unicode' to an encoded utf-8 byte string '''
    if isinstance(obj, unicode):
        obj = obj.encode('utf-8', errors)
    return obj


class HawkIOException(Exception):
    """Base Hawk Error. All other Errors thrown by hawk should inherit from this. """

    def __init__(self, value=None):
        # super(Exception).__init__(self)
        self.value = value

    def __str__(self):
        return "%s" % to_utf8(self.value)

    def __unicode__(self):
        return "%s" % to_unicode(self.value)


class HawkBaseError(HawkIOException):
    def __init__(self, msg, errors=None):
        HawkIOException.__init__(self, msg)


class HawkCoreError(HawkIOException):
    """Class to catch hawk core errors. """

    def __init__(self, msg, errors=None, api_server=None):
        HawkBaseError.__init__(self, msg)
        if api_server is not None:
            self.api_server = api_server
        self.errors = errors


class HawkApiError(HawkIOException):
    """Class to catch hawk api errors. """

    def __init__(self, msg, errors=None):
        HawkBaseError.__init__(self, msg)
        self.errors = errors

# TODO Finish validation vars.


class HawkApiValidationError(HawkApiError):
    """Pre-check if json is valid, If not raise this error. """

    def __init__(self, msg, errors=None, ):
        HawkApiError.__init__(self, msg, errors, )
        self.something = something


class HawkApiIncidentAddError(HawkApiError):
    """Class to catch error adding incident to api. """

    def __init__(self, msg, errors=None):
        HawkApiError.__init__(self, msg, errors)


class HawkLoginError(HawkIOException):
    """ This class means Login to API failed. """

    def __init__(self, msg, errors=None, user=None):
        HawkBaseError.__init__(self, msg)
        if user is not None:
            self.user = user
        self.errors = errors


class HawkLogoutError(HawkIOException):
    """ This class means Login to API failed. """

    def __init__(self, msg, errors=None, user=None):
        HawkBaseError.__init__(self, msg)
        if user is not None:
            self.user = user
        self.errors = errors




import os
import sys
from ctypes import cdll, byref, create_string_buffer


def hawk_print(val):
    if not sys.stderr.closed:
        sys.stderr.write("%s\n" % val)


def set_proc_name(newname):
    # Linux-only best effort; noop on non-POSIX platforms (e.g. Windows).
    if os.name != "posix":
        return False
    try:
        libc = cdll.LoadLibrary('libc.so.6')
        buff = create_string_buffer(len(newname)+1)
        buff.value = newname
        libc.prctl(15, byref(buff), 0, 0, 0)
        return True
    except Exception:
        return False



import logging
import os
from datetime import datetime, timedelta, timezone

from base import BaseModel


"""
location_address.(location)
location_region.(location)
location_point.(location)

it_asset.(asset)
data.(asset)
organization.(asset)
person.(asset)
circuit.(it_asset)
computing_device.(it_asset)
database.(it_asset)
network.(it_asset)
service.(it_asset)
software.(it_asset)
system.(it_asset)
website.(it_asset)
"""

# Custom Data types
'''
CidrType:
    name: cidr-type
    type: String
    serializable
    pattern: ([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))\.([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))\.([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))\.([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))/([0-9]|[1-2][0-9]|3[0-2])
    Notes: use python library ipaddress for this.

CpeType:
    name: CpeType
    type: String
    serializable
    name: cpe-type
    memberTypes:
        name: Cpe23Type
            url: http://cpe.mitre.org/naming/2.0#cpe23Type
            type: String
            DocString: Define the format for acceptable CPE Names. A string format is used with
                the id starting with the word cpe:2.3 followed by : and then some number of individual components
                separated by colons.
            pattern: cpe:2\.3:[aho\*\-](:(((\?*|\*?)([a-zA-Z0-9\-\._]|(\\[\\\*\?!"#$$%25&'\(\)\+,/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){5}(:(([a-zA-Z]{2,3}(-([a-zA-Z]{2}|[0-9]{3}))?)|[\*\-]))(:(((\?*|\*?)([a-zA-Z0-9\-\._]|(\\[\\\*\?!"#$$%25&'\(\)\+,/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){4}

        name: Cpe22Type
            url: http://cpe.mitre.org/naming/2.0#cpe22Type
            type: String
            DocString: Define the format for acceptable CPE Names. A URN format is used with the
                id starting with the word cpe followed by :/ and then some number of individual components separated by
                colons.
            pattern: [cC][pP][eE]:/[AHOaho]?(:[A-Za-z0-9\._\-~%25]*){0,6}


    Notes: Common Platform Enumeration now being ran by NIST


HostnameType:
        name: HostnameType
        type: String
        serializable
        name: hostname-type
        baseType: token http://www.eclipse.org/emf/2003/XMLType#token
        pattern: [\w\-]+(\.[\w\-]+){0,}

Ipv4Type:
        name: Ipv4Type
        type: String
        serializable
        name: ipv4-type
        baseType: Token
        pattern: ([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))\.([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))\.([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))\.([0-9]|[1-9][0-9]|1([0-9][0-9])|2([0-4][0-9]|5[0-5]))

Ipv6Type
        name: Ipv6Type
        type: String
        serializable
        name: ipv6-type
        baseType: Token
        pattern: ([0-9a-fA-F]{1,4}:){7}[0-9a-fA-F]{1,4}

LatitudeType
        name: LatitudeType
        type: double
        serializable
        name: latitude_._type
        minInclusive: -90
        maxInclusive: 90

LatitudeTypeObject
        name: LatitudeTypeObject
        type: double
        serializable
        name: latitude_._type:Object
        baseType: latitude_._type

LongitudeType
        name: LongitudeType
        type: double
        serializable
        name: longitude_._type
        minExclusive: -180
        maxInclusive: 180

LongitudeTypeObject
        name: LongitudeTypeObject
        type: double
        serializable
        name: longitude_._type:Object
        baseType: longitude_._type

LocaleType:
        name: LocaleType
        type: String
        serializable
        name: locale-type
        baseType: Token
        pattern: [a-zA-Z]{2,3}(-([a-zA-Z]{2}|[0-9]{3}))?

MacAddressType
        name: MacAddressType
        type: String
        serializable
        name: mac-address-type
        baseType: Token
        pattern: ([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}

PortType
        name: PortType
        type: BigInteger
        serializable
        name: port-type
        baseType: http://www.eclipse.org/emf/2003/XMLType#integer
        minInclusive: 0
        maxInclusive: 65535

RadiusType
        name: RadiusType
        type: double
        serializable
        name: radius_._type
        minInclusive: 0

RadiusTypeObject
        name: RadiusTypeObject
        type: double
        serializable
        name: radius_._type:Object
        baseType: radius_._type

TelephoneNumberType
        name: TelephoneNumberType
        type: String
        serializable
        name: telephone-number-type
        pattern: (([2-9][0-8]\d-[2-9]\d{2}-[0-9]{4})|(\+([0-9]%20?){6,14}[0-9]))

'''


class AssetIdentificationType(object):
    '''
        class:
            name: asset-identification-type
            kind: elementOnly
            Super Types: AssetsType -> RelationshipsContainerType

        Attribute:
            Name: assetRef
            Lower bound = 1
            Upper bound = 1
            ordered, unique, changeable
            Type: String
    '''
    pass


class AssetsType(object):
    '''
        class:
            name: assets-type
            kind: elementOnly
            Super Types: RelationshipsContainerType
        Reference:
            name: asset
            Lower bound = 1
            upper bound = *
            Containment, ordered, unique, changeable
            Type: AssetType _obj
            Kind: element
            namespace: ##targetNamespace
    '''
    pass


class AssetType(object):
    '''
        class:
            name: asset_._type
            kind: elementOnly
        Attribute:
            name: assetGroup
            Lower bound = 1
            Upper bound = 1
            ordered, unique, changeable
            DocString: Holds identifying attributes that are common to all assets.
            Type: FeatureMapEntry??
            Kind: group (list?)
            namespace: ##targetNamespace
        Reference:
            name: asset
            Lower = 1
            Upper = 1
            Containment, transient, derived, ordered, unique, volatile
            DocString: Holds identifying attributes that are common to all assets.
            kind: element
            namespace: ##targetNamespace
            group: assetGroup
            type: AssetType1
        Attribute:
            name: id
            type: NCName | String
            Lower = 1
            Upper = 1
            ordered, unique, changeable
            DocString: An internal ID to identify this asset.
            kind: Attribute
        Attribute:
            name: anyAttribute
            type: FeatureMapEntry
            Lower = 0
            Upper = *
            ordered, changeable
            DocString: A placeholder so that content creators can add attributes as desired.
            kind: attributeWildcard
            wildcards: ##other
            processing: strict
    '''
    pass


class AssetType1(object):
    '''
        class:
            name: AssetType1
            Abstract (weather instances of this class can be created)
            name: asset-type
            kind: elementOnly
        Reference:
            name: syntheticId
            type: SyntheticIdType
            Lower = 0
            Upper = *
            Containment, ordered, unique, changeable
            DocString: Holds the synthetic ID for the asset
            kind: element
            name: synthetic-id
            namespace: ##targetNamespace
        Reference:
            name: locations
            type: LocationsType
            lower = 0
            Upper = 1
            Containment, ordered, unique, changeable
            DocString: One or more locations where this asset resides
            kind: element
            name: locations
            namespace: ##targetNamespace
        Reference:
            name: extendedInformation
            type: ExtendedInformationType
            lower = 0
            uppoer = 1
            Containment, ordered, unique, changeable
            DocString: This is a container to hold any addtional identifying information for an asset, as specified by the content creator.
            kind: element
            name: extended-information
            namespace: ##targetNamespace
        Attribute:
            name: timestamp
            type: datetime.now(tz=timezone.utc)
            lower = 0
            uppoer = 1
            ordered, unique, changeable
            DocString: Indicates when the data was last known to be correct.
            kind: attribute
            name: timestamp
            namespace: ##targetNamespace
    '''
    pass


class ExtendedInformationType(object):
    '''
        class:
            name: ExtendedInformationType
            kind: elementOnly
        Attribute:
            name: any
            type: featureMapEntry
            lower 1
            upper *
            kind: elementWildcard
            wildcards: ##other
            processing: lax
    '''
    pass


class BirthdateType(object):
    '''
        class:
            name: BirthdateType
            name: birthdate_._type
            kind: simple
        Attribute:
            name: value
            type: Date
            lower = 0
            uppoer = 1
            ordered, unique, changeable
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            ordered, unique, changeable
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
            namespace: ##targetNamespace
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for date.
            # TODO: convert to datetime object.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class CidrType1(object):
    '''
        class:
            name: CidrType1
            kind: simple
        Attribute:
            name: value
            Type: CidrType
            lower = 0
            uppoer = 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            uppoer = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            Attribute: timestamp
            Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for CidrType
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class CircuitNameType(object):
    '''
        class:
            name: CircuitNameType
            kind: simple
        Attribute:
            name: value
            Type: Token (String)
            lower = 0
            uppoer = 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            uppoer = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            Attribute: timestamp
            Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class CircuitType(object):
    '''
        class:
            name: CircuitType
            Super Type: itAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: circuitName
            Type: CircuitNameType
            lower = 0
            uppoer = 1
            DocString: The common name for the circult.
            kind: element
    '''
    pass


class ComputingDeviceType(object):
    '''
        class:
            name: ComputingDeviceType
            Super Types: itAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: distinguishedName
            type: DistinguishedNameType
            lower = 0
            upper = 1
            DocString: The full X.500 distinguished name for the device.
            kind: element
        Reference:
            name: cpe
            type: CpeType1
            lower = 0
            uppoer = *
            DocString: The hardware CPE name for the device (CPE 2.2 URI or CPE 2.3 Formatted String).
            kind: element
        Reference:
            name: connections
            type: ConnectionsType
            lower 0
            uppoer 1
            DocString: The IP network interface connections that exist for the device (regardless
                                of if the network interface is connected to a network or not).
            kind: element
        Reference:
            name: fqdn
            Type: FqdnType
            lower 0
            upper 1
            DocString: The fully qualified domain name for the object.
            kind: element
        Reference:
            name: hostname
            type: HostnameType1
            lower 0
            upper 1
            DocString: The hostname of the computing device.
            kind: element
        Reference:
            name: motherboardGuid
            type: MotherboardGuidType
            lower 0
            upper 1
            DocString: The motherboard globally unique identifier of the computing device.
            kind: element
    '''
    pass


class ConnectionsType(object):
    '''
        class:
            name: ConnectionsType
            kind: elementOnly
        Reference:
            name: connection
            type: NetworkInterfaceType
            lower: 1
            upper *
            DocString: An IP network interface connection.
            kind: element
    '''
    pass


class CpeType1(object):
    '''
        class:
            name: CpeType1
            kind: simple
        Attribute:
            name: value
            type: CpeType
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for CpeType
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class DatabaseType(object):
    '''
        class:
            name: DatabaseType
            Super Types: itAssetType -> AssetType1
            kind elementOnly
        Reference:
            name: instanceName
            type: InstanceNameType
            lower 0
            upper 1
            DocString: The name of the database instance being identified.
            kind: element
    '''
    pass


class DataType(object):
    '''
        class:
            name: DataType
            Suepr Types: AssetType1
            kind: elementOnly
    '''
    pass


class DistinguishedNameType(object):
    '''
        class:
            name: DistinguishedNameType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class EmailAddressType(object):
    '''
        class:
            name: EmailAddressType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class FqdnType(object):
    '''
        class:
            name: FqdnType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class HostnameType1(object):
    '''
        class:
            name: HostnameType1
            kind: simple
        Attribute:
            name: value
            type: HostnameType (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for HostnameType
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class HostType(object):
    '''
        class:
            name: HostType
            kind: elementOnly
        Reference:
            name: fqdn
            type: FqdnType
            lower 0
            upper 1
            DocString: The fully qualified domain name for the object.
            kind: element
        Reference:
            name: ipAddress
            type: IpAddressType
            lower 0
            upper 1
            DocString: An IP address
            kind: element
    '''
    pass


class InstallationIdType(object):
    '''
        class:
            name: InstallationIdType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class InstanceNameType(object):
    '''
        class:
            name: InstanceNameType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class IpAddressType(object):
    '''
        class:
            name: IpAddressType
            kind: elementOnly
        Reference:
            name: ipV4
            type: IpV4Type1
            lower 0
            upper 1
            kind: element
        Reference:
            name: ipV6
            type: IpV6Type1
            kind: element
    '''
    pass


class IpNetRangeType(object):
    '''
        class:
            name: IpNetRangeType
            kind: elementOnly
        Reference:
            name: ipNetRangeStart
            type: IpAddressType
            lower 1
            upper 1
            DocString: The starting IP address in the network.
            kind: element
        Reference:
            name: ipNetRangeEnd
            type: IpAddressType
            lower 1
            upper 1
            DocString: The ending IP address in the network.
            kind: element
    '''
    pass


class IpV4Type1(object):
    '''
        class:
            name: IpV4Type1
            kind: simple
        Attribute:
            name: value
            type: Ipv4Type (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for IpV4Type
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class IpV6Type1(object):
    '''
        class:
            name: IpV6Type1
            kind: simple
        Attribute:
            name: value
            type: Ipv6Type (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for ipv6 type
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class ItAssetType(object):
    '''
        class:
            name: ItAssetType
            Super Type: AssetType1
            kind: elementOnly
    '''
    pass


class LicenseType(object):
    '''
        class:
            name: LicenseType
            kind: simple
        Attribute:
            name: value
            type: (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class LocaleType1(object):
    '''
        class:
            name: LocaleType1
            kind: simple
        Attribute:
            name: value
            type: LocaleType (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for LocaleType
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class LocationPointType(object):
    '''
        class:
            name: LocationPointType
            kind: empty
        Attribute:
            name: elevation
            type: double
            lower 0
            upper 1
            DocString: The elevation of the asset, specified in meters from sea level.
            kind: attribute
        Attribute:
            name: latitude
            type: LatitudeType
            lower 1
            upper 1
            DocString: The latitude of the asset, defined between -90 (90 degrees South, inclusive) and 90 (90 degrees North, inclusive).
            kind: attribute
        Attribute:
            name: longitude
            type: LongitudeType
            lower 1
            upper 1
            DocString: The longitude of the asset, defined between -180 (180 degrees West, exclusive) and 180 (180 degrees East, inclusive).
            kind: attribute
        Attribute:
            name: radius
            type: RadiusType
            lower 0
            upper 1
            DocString: The radius of a horizontal circle centered on the point within which the asset resides.
            kind: attribute
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''
    pass


class LocationRegionType(object):
    '''
        class:
            name: LocationRegionType
            kind: simple
        Attribute:
            name: value
            type: NormalizedString (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for NormalizedString
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class LocationsType(object):
    '''
        class:
            name: LocationsType
            kind: elementOnly
        Attribute:
            name: locationGroup
            type: FeatureMapEntry
            lower 0
            upper *
            DocString: The base for a subtitution group for elements that contain location information.
            kind: group
            name: location:group
        Reference:
            name: location
            type: EObject [org.eclipse.emf.ecore.EObject]
            lower 1
            upper *
            DocString: The base for a subtitution group for elements that contain location information.
            kind: element
            group: location:group

    '''
    pass


class MacAddressType1(object):
    '''
        class:
            name: MacAddressType1
            kind: simple
        Attribute:
            name: value
            type: MacAddressType (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for MacAddressType
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class MotherboardGuidType(object):
    '''
        class:
            name: MotherboardGuidType
            kind: simple
        Attribute:
            name: value
            type: String (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class NetworkInterfaceType(object):
    '''
        class:
            name: NetworkInterfaceType
            kind: elementOnly
        Reference:
            name: ipAddress
            type: IpAddressType
            lower 0
            upper 1
            DocString: The IP address for this network interface.
            kind: element
        Reference:
            name: macAddress
            type: MacAddressType1
            lower 0
            upper 1
            DocString: The MAC address associated with this network interface.
        Reference:
            name: url
            type: UrlType
            lower 0
            upper 1
            DocString: A URL in a relevant DNS server for this IP address.
            kind: element
        Reference:
            name: subnetMask
            type: IpAddressType
            lower 0
            upper 1
            DocString: The subnet mask address for this network interface.
            kind: element
        Reference:
            name: defaultRoute
            type: IpAddressType
            lower 0
            upper 1
            DocString: The IP address for the default gateway of this connection.
            kind: element
    '''
    pass


class NetworkNameType(object):
    '''
        class:
            name: NetworkNameType
            kind: simple
        Attribute:
            name: value
            type: NormalizedString (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: Add validation for NormalizedString.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class NetworkType(object):
    '''
        class:
            name: NetworkType
            Super Types: ItAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: networkName
            type: NetworkNameType
            lower 0
            upper 1
            DocString: The name of the network as commonly referred to.
            kind: element
        Reference:
            name: ipNetRange
            type: IpNetRangeType
            lower 0
            upper 1
            DocString: The start and end IP addresses to indicate the range of IP addresses covered by this network.
            kind: element
        Reference:
            name: cidr
            type: CidrType1
            lower 0
            upper 1
            DocString: The classless inter-domain routing notation for the network.
            kind: element
    '''
    pass


class OrganizationType(object):
    '''
        class:
            name: OrganizationType
            Super Types: AssetType1
            kind: elementOnly
        Reference:
            name: organisationNameDetails
            type: OrganisationNameDetailsType -> OrganisationNameDetails
            lower 0
            upper *
            DocString: A container for organisation name details.
            kind: element
            namespace: urn:oasis:names:tc:ciq:xsdschema:xNL:2.0
        Reference:
            name: emailAddress
            type: EmailAddressType
            lower 0
            upper *
            DocString: An email address
            kind: element
        Reference:
            name: telephoneNumber
            type: TelephoneNumberType1
            lower 0
            upper *
            DocString: The telephone number. For a North American number, the number must be valid and the format
                must be XXX-XXX-XXXX where X is a digit. For an international number, the number must begin with a '+'
                symbol, followed by 7 to 15 digits. A space may be used between digits, as appropriate. For example: +88
                888 888 8 (this is following the ITU-T E.123 notation).
            kind: element
        Reference:
            name: websiteUrl
            type: WebsiteUrlType
            lower 0
            upper *
            DocString: The URL to the website.
            kind: element
    '''
    pass


class PersonType(object):
    '''
        class:
            name: PersonType
            Super Types: AssetType1
            kind: elementOnly
        Reference:
            name: personName
            type: PersonNameType -> PersonName
            lower 0
            upper 1
            DocString: Container for person name details.
            kind: element
            namespace: urn:oasis:names:tc:ciq:xsdschema:xNL:2.0
        Reference:
            name: emailAddress
            type: EmailAddressType
            lower 0
            upper *
            DocString: An email address
            kind: element
        Reference:
            name: telephoneNumber
            type: TelephoneNumberType1
            lower 0
            upper *
            DocString: The telephone number. For a North American number, the number must be valid and the format
                must be XXX-XXX-XXXX where X is a digit. For an international number, the number must begin with a '+'
                symbol, followed by 7 to 15 digits. A space may be used between digits, as appropriate. For example: +88
                888 888 8 (this is following the ITU-T E.123 notation).
            kind: element
        Reference:
            name: birthdate
            type: BirthdateType
            lower 0
            upper 1
            DocString: The birthdate of the person.
            kind: element
    '''
    pass


class PortRangeType(object):
    '''
        class:
            name: PortRangeType
            kind: empty
        Attribute:
            name: lowerBound
            type: PortType (BigInteger)
            lower 1
            upper 1
            kind: attribute
        Attribute:
            name: upperBound
            type: PortType (BigInteger)
            lower 1
            upper 1
            kind: attribute
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''
    pass


class PortType1(object):
    '''
        class:
            name: PortType1
            kind: simple
        Attribute:
            name: value
            type: PortType (BigInteger)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise
        except:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = int

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            if val <= 0 and val <= 65535:
                return val
            else:
                raise ValueError(
                    "Expected int between 0 and 65535, received {}".format(val))
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class ProtocolType(object):
    '''
        class:
            name: ProtocolType
            kind: simple
        Attribute:
            name: value
            type: String
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class ServiceType(object):
    '''
        class:
            name: ServiceType
            Super Types: ItAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: host
            type: HostType
            lower 0
            upper 1
            DocString: The hostname or IP address where the service is hosted.
            kind: element
        Reference:
            name: port
            type: PortType1
            lower 0
            upper *
            DocString: The port to which the service is bound.
            kind: element
        Reference:
            name: portRange
            type: PortRangeType
            lower 0
            upper *
            DocString: The inclusive port range to which the service is bound.
            kind: element
        Reference:
            name: protocol
            type: ProtocolType
            lower 0
            upper 1
            DocString: The protocol used to interact with the service.
            kind: element
    '''
    pass


class SoftwareType(object):
    '''
        class:
            name: SoftwareType
            Super Types: ItAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: installationId
            type: InstallationIdType
            lower 0
            upper 1
            DocString: Any identifier for the software instance (installation)
            kind: element
        Reference:
            name: cpe
            type: CpeType1
            lower 0
            upper 1
            DocString: The CPE name for the software (CPE 2.2 URI or CPE 2.3 Formatted String).
            kind: element
        Reference:
            name: license
            type: LicenseType
            lower 0
            upper *
            DocString: The license key for the software.
            kind: element
    '''
    pass


class SyntheticIdType(object):
    '''
        class:
            name: SyntheticIdType
            kind: empty
        Attribute:
            name: id
            type: Token
            lower 1
            upper 1
            DocString: The ID of the asset within the resource namespace.
            kind: attribute
        Attribute:
            name: resource
            type: AnyURI [java.lang.String]
            lower 1
            upper 1
            DocString: The namespace governing this synthetic ID.
            kind: attribute
    '''
    pass


class SystemNameType(object):
    '''
        class:
            name: SystemNameType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class SystemType(object):
    '''
        class:
            name: SystemType
            Super types: ItAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: systemName
            type: SystemNameType
            lower 0
            upper *
            DocString: The name of the system. It is possible that a system have multiple names, or even abbreviated names.
                        Each one of those names may be captured here.
            kind: element
        Reference:
            name: version
            type: VersionType
            lower 0
            upper 1
            DocString: The version of the system.
            kind: element
    '''
    pass


class TelephoneNumberType1(object):
    '''
        class:
            name: TelephoneNumberType1
            kind: simple
        Attribute:
            name: value
            type: TelephoneNumberType (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            # TODO: add validation for phone number
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class UrlType(object):
    '''
        class:
            name: UrlType
            kind: simple
        Attribute:
            name: value
            type: AnyURI (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class VersionType(object):
    '''
        class:
            name: VersionType
            kind: simple
        Attribute:
            name: value
            type: Token (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class WebsiteType(ItAssetType):
    '''
        class:
            name: WebsiteType
            Super Types: ItAssetType -> AssetType1
            kind: elementOnly
        Reference:
            name: documentRoot??
            type: DocumentRootType??
            lower 0
            upper 1
            DocString: The absolute path to the document root location of the website on the host.
            kind: element
        Reference:
            name: locale
            type: LocaleType1
            lower 0
            uppoer 1
            DocString: The locale of the website represented as an RFC 5646 language, and optionally, region code.
            kind: element
    '''
    pass


class WebsiteUrlType(object):
    '''
        class:
            name: WebsiteUrlType
            kind: simple
        Attribute:
            name: value
            type: AnyURI (String)
            lower 0
            upper 1
            kind: simple
        Attribute:
            name: source
            type: String
            lower = 0
            upper = 1
            DocString: Contains the source of the information. The value of this field is left open to the
                content producer, but MAY include a synthetic ID of the asset which sourced the information, another ID
                of the source, or a description of the source.
            kind: attribute
            name: source
        Attribute: timestamp
        Attribute: anyAttribute (list of AnyAttribute() objects)
    '''

    def __init__(self, attributes=None):
        self._value = ''
        self._source = ''
        self.timestamp = None
        self.attributes = attributes

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise

        self._timestamp()

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, src):
        try:
            self._source = self._source_validate(src)
        except TypeError:
            raise

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''
        TYPE = str

        if isinstance(val, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))

    def _source_validate(self, src):
        ''' Internal function to validate source attribute. '''
        TYPE = str

        if isinstance(src, TYPE):
            return src
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(src).__name__))


class AnyAttribute(object):
    '''A placeholder so that content creators can add attributes as desired.
        Args:
            name: (str) Name of attribute.
            value: (str, dict, list, tuple) Value can be any built-in data type.

    '''

    def __init__(self, name, value, source):
        self._name = name
        self._value = value
        self._value_type = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        try:
            self._name = self._name_validate(n)
        except TypeError:
            raise

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        try:
            self._set_value_type(val)
        except:
            raise

        try:
            self._value = self._value_validate(val)
        except TypeError:
            raise
        except:
            raise

        self._timestamp()

    def _set_value_type(val):
        '''Verifies value is builtin data type and sets value type.'''
        if isinstance(val, (str, list, dict, int, tuple, bool)):
            self._value_type = type(val)
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(n).__name__))

    def _timestamp(self):
        ''' Internal function to update timestamp when value is set. '''
        self.timestamp = datetime.now(timezone.utc)

    def _name_validate(self, n):
        ''' Internal function to validate name attribute. '''

        TYPE = str

        if isinstance(n, TYPE):
            # Other validation here if needed. Raise exceptions as needed.
            return n
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(n).__name__))

    def _value_validate(self, val):
        ''' Internal function to validate value attribute. '''

        if isinstance(val, self._value_type):
            return val
        else:
            raise TypeError("Expected '{}' but received a '{}'".format(
                TYPE.__name__, type(val).__name__))


'''
class location(BaseModel):

class locationAddress(location):

class locationPoint(location):




class asset(BaseModel):

class itAsset(asset):

class data(asset):

class organization(asset):

class person(asset):

class circuit(itAsset):

class computingDevice(itAsset):

class database(itAsset):

class network(itAsset):

class service(itAsset):

class software(itAsset):

class system(itAsset):

class website(itAsset):
'''




try:
    import configparser  # Python3
except ImportError:
    import ConfigParser as configparser  # Python2.7


class hawkParseConfig(object):

    def __init__(self, cfg_file):
        self.cfg_file = cfg_file
        cfg = configparser.RawConfigParser()
        logger.info("Parsing Configuration: %s" % (cfg))
        fp = open(self.cfg_file)

        self.config = self.cfg.readfp(fp)
        self.self.ret = {}


def parseConfig(cfg):

    try:
        config = ConfigParser.ConfigParser()
        fp = open(cfg)
        config.readfp(fp)
        self.ret['server'] = config.get("API", "server")
        if type(self.ret['server']) is str:
            if "," in self.ret['server']:
                self.ret['server'] = self.ret['server'].split(',')
            else:
                self.ret['server'] = [self.ret['server']]

        self.ret['username'] = config.get("API", "username")
        self.ret['password'] = config.get("API", "password")
        self.ret['insecure'] = config.get("SETTINGS", "insecure")
        self.ret['global'] = config.get("SETTINGS", "global")
        self.ret['send_emails'] = config.get("SETTINGS", "send_emails")
        self.ret['timeout'] = config.get("SETTINGS", "timeout")
        self.ret['retry'] = config.get("SETTINGS", "retry")

        self.ret['smtp_from'] = config.get("SETTINGS", "smtp_from")

        try:
            self.ret['proxy_type'] = config.get("SETTINGS", "proxy_type")
            self.ret['proxy_host'] = config.get("SETTINGS", "proxy_host")
            self.ret['proxy'] = True
        except:
            self.ret['proxy'] = False
            pass

        fp.close()

    except Exception, e:
        hawk_print("Failed to parse config %s: [%s]" % (cfg, e))
        sys.exit()
    return self.ret



import json
import os
import tempfile
import unittest

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa

from hawkSoar.security.hybrid_catalog import HybridToolCatalog
from hawkSoar.security.hybrid_execution import HybridTaskExecutor


def _base_task():
    return {
        "task_id": "11111111-1111-1111-1111-111111111111",
        "group_id": "22222222-2222-2222-2222-222222222222",
        "case_id": "33333333-3333-3333-3333-333333333333",
        "category": "NETWORK",
        "action": "ping",
        "parameters": {"target": "127.0.0.1", "count": 1},
        "requires_approval": True,
        "timeout": 5,
        "priority": "LOW",
    }


class HybridExecutionTest(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        public_key = self.private_key.public_key().public_bytes(
            serialization.Encoding.PEM,
            serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        self.pub_path = os.path.join(self.temp_dir.name, "public.pem")
        with open(self.pub_path, "wb") as handle:
            handle.write(public_key)

    def tearDown(self):
        self.temp_dir.cleanup()

    def _sign(self, task):
        canonical = json.dumps(task, sort_keys=True, separators=(",", ":")).encode("utf-8")
        sig = self.private_key.sign(canonical, padding.PKCS1v15(), hashes.SHA256())
        return sig.hex()

    def _executor(self, expected_group_getter=None, **overrides):
        config = {
            "task_signing_public_key": self.pub_path,
            "enforce_signed_hybrid_tasks": True,
            "enforce_group_id": True,
            "max_task_timeout": 60,
            "plugin_data_path": self.temp_dir.name,
            "hybrid_dynamic_tools_file": os.path.join(self.temp_dir.name, "hybrid_dynamic_tools.json"),
            "enable_command_execution": True,
        }
        config.update(overrides)

        def fake_runner(cmd, timeout):
            del timeout
            return 0, "ok: %s" % " ".join(cmd), ""

        return HybridTaskExecutor(
            config=config,
            expected_group_id_getter=expected_group_getter
            or (lambda: "22222222-2222-2222-2222-222222222222"),
            command_runner=fake_runner,
        )

    def test_blocks_missing_signature(self):
        executor = self._executor()
        response = executor.process(_base_task(), signature=None)
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("signature", response["error"].lower())

    def test_blocks_non_object_task_payload(self):
        executor = self._executor()
        response = executor.process(None, signature=None)
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("json object", response["error"].lower())

    def test_blocks_group_mismatch(self):
        executor = self._executor()
        task = _base_task()
        task["group_id"] = "aaaaaaaa-2222-2222-2222-222222222222"
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("group_id", response["error"])

    def test_blocks_disallowed_wmic(self):
        executor = self._executor()
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "wmic_query"
        task["parameters"] = {"query": "process_create"}
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("allowlisted", response["error"])

    def test_blocks_containment_without_approval(self):
        executor = self._executor()
        task = _base_task()
        task["category"] = "CONTAINMENT"
        task["action"] = "kill_process"
        task["requires_approval"] = False
        task["parameters"] = {"pid": 100}
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("require approval", response["error"].lower())

    def test_successful_signed_task(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "SUCCESS")
        self.assertIn("stdout", response["output"])

    def test_blocks_remote_process_snapshot_when_policy_disabled(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_process_snapshot"
        task["requires_approval"] = True
        task["parameters"] = {
            "target_host": "host01",
            "pid": 1234,
            "username": "admin",
            "password": "pw",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("disabled by policy", response["error"].lower())

    def test_blocks_remote_process_snapshot_by_name_without_approval(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_process_snapshot_by_name"
        task["requires_approval"] = False
        task["parameters"] = {
            "target_host": "host01",
            "process_name": "lsass",
            "username": "admin",
            "password": "pw",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("requires approval", response["error"].lower())

    def test_blocks_remote_file_collect_when_policy_disabled(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_file_collect"
        task["requires_approval"] = True
        task["parameters"] = {
            "target_host": "host01",
            "remote_path": "C:\\Windows\\Temp\\a.bin",
            "username": "admin",
            "password": "pw",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("disabled by policy", response["error"].lower())

    def test_blocks_run_command_without_approval(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "run_command"
        task["requires_approval"] = False
        task["parameters"] = {"argv": ["whoami"]}
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("requires approval", response["error"].lower())

    def test_register_and_execute_dynamic_tool(self):
        executor = self._executor(enforce_group_id=False)
        register = _base_task()
        register["category"] = "ENDPOINT"
        register["action"] = "register_dynamic_tool"
        register["requires_approval"] = True
        register["parameters"] = {
            "tool_id": "custom.echo",
            "name": "Custom Echo",
            "category": "ENDPOINT",
            "action": "custom_echo",
            "description": "Echo a message.",
            "command": ["echo", "{message}"],
            "parameters_schema": {"message": "hello"},
            "requires_approval": True,
            "risk": "HIGH",
        }
        register_response = executor.process(register, signature=self._sign(register))
        self.assertEqual(register_response["status"], "SUCCESS")

        run_custom = _base_task()
        run_custom["category"] = "ENDPOINT"
        run_custom["action"] = "custom_echo"
        run_custom["requires_approval"] = True
        run_custom["parameters"] = {"message": "hello"}
        run_response = executor.process(run_custom, signature=self._sign(run_custom))
        self.assertEqual(run_response["status"], "SUCCESS")
        self.assertIn("stdout", run_response["output"])

    def test_blocks_remote_command_execute_when_policy_disabled(self):
        executor = self._executor(enforce_group_id=False, enable_remote_command_execution=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_command_execute"
        task["requires_approval"] = True
        task["parameters"] = {
            "target_host": "host01",
            "username": "admin",
            "password": "pw",
            "argv": ["whoami"],
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")

    def test_blocks_remote_registry_query_when_policy_disabled(self):
        executor = self._executor(enforce_group_id=False, enable_remote_registry_read=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_registry_query"
        task["requires_approval"] = False
        task["parameters"] = {
            "target_host": "host01",
            "path": "HKLM\\SOFTWARE",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")

    def test_blocks_register_windows_service_without_approval(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "register_windows_service"
        task["requires_approval"] = False
        task["parameters"] = {
            "service_name": "HawkSoard",
            "executable_path": "C:\\does-not-exist\\hawk-soard.exe",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("requires approval", response["error"].lower())

    def test_blocks_restart_windows_service_without_approval(self):
        executor = self._executor(enforce_group_id=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "restart_windows_service"
        task["requires_approval"] = False
        task["parameters"] = {"service_name": "HawkSoard"}
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("requires approval", response["error"].lower())

    def test_catalog_executable_actions_are_wired(self):
        executor = self._executor(enforce_group_id=False)
        catalog = HybridToolCatalog(config=executor.config).export(group_id="g-1")
        executable_actions = set(
            tool["action"]
            for tool in catalog["tools"]
            if tool.get("execution_mode") == "executable" and isinstance(tool.get("action"), str)
        )
        allowlisted = set()
        for actions in executor.ALLOWED_ACTIONS.values():
            allowlisted.update(actions)
        missing_allowlist = sorted(x for x in executable_actions if x not in allowlisted)
        missing_handlers = sorted(
            x
            for x in executable_actions
            if x not in executor._action_handlers and x not in executor.dynamic_tool_actions
        )
        self.assertEqual([], missing_allowlist)
        self.assertEqual([], missing_handlers)

    def test_blocks_remote_config_update_without_approval(self):
        executor = self._executor(enforce_group_id=False, enable_remote_config_management=True)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_config_update"
        task["requires_approval"] = False
        task["parameters"] = {
            "target_host": "host01",
            "remote_path": "C:\\ProgramData\\Hawk\\hawk-soard.cfg",
            "content": "[SETTINGS]\n",
            "username": "admin",
            "password": "pw",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("requires approval", response["error"].lower())

    def test_blocks_remote_service_restart_when_policy_disabled(self):
        executor = self._executor(enforce_group_id=False, enable_remote_config_management=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_service_restart"
        task["requires_approval"] = True
        task["parameters"] = {
            "target_host": "host01",
            "service_name": "HawkSoard",
            "username": "admin",
            "password": "pw",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")

    def test_blocks_remote_process_restart_when_policy_disabled(self):
        executor = self._executor(enforce_group_id=False, enable_remote_config_management=False)
        task = _base_task()
        task["category"] = "ENDPOINT"
        task["action"] = "remote_process_restart"
        task["requires_approval"] = True
        task["parameters"] = {
            "target_host": "host01",
            "process_name": "hawk-soard",
            "executable_path": "C:\\Program Files\\Hawk\\hawk-soard.exe",
            "arguments": ["-c", "C:\\ProgramData\\Hawk\\hawk-soard.cfg"],
            "username": "admin",
            "password": "pw",
        }
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")

    def test_blocks_when_local_group_unavailable(self):
        executor = self._executor(expected_group_getter=lambda: None)
        task = _base_task()
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("local node group is unavailable", response["error"].lower())

    def test_blocks_when_group_getter_missing(self):
        executor = HybridTaskExecutor(
            config={
                "task_signing_public_key": self.pub_path,
                "enforce_signed_hybrid_tasks": True,
                "enforce_group_id": True,
                "max_task_timeout": 60,
            },
            expected_group_id_getter=None,
            command_runner=lambda cmd, timeout: (0, "ok", ""),
        )
        task = _base_task()
        response = executor.process(task, signature=self._sign(task))
        self.assertEqual(response["status"], "BLOCKED")
        self.assertIn("group resolver", response["error"].lower())


if __name__ == "__main__":
    unittest.main()




import time
import random
import websocket
import ssl
import json
import threading
from multiprocessing import Queue
from uuid import uuid4
import socket
import psutil

from .auth import Authentication
from hawkSoar.common.logger import get_logger

# for connecting to soar API


class soarAPI(threading.Thread):
    """Communicating with the SOAR API. """
    config = {}
    authObj = None
    uuid = None
    isRunning = False
    ws = None
    incoming_queue = None
    outgoing_queue = None

    def __init__(self, config):
        threading.Thread.__init__(self)
        self.config = config
        self.logging = get_logger()
        self.incoming_queue = Queue()
        self.outgoing_queue = Queue()
        self._sender_thread = None

        try:
            self.authObj = Authentication(random.choice(
                self.config['server']), self.config['clientId'], self.config['clientSecret'])
            if not self.config['insecure']:
                self.authObj.setVerify(True)
            else:
                self.authObj.setVerify(False)

        except Exception as e:
            self.logging.error(str(e))
            raise e

        try:
            if config['client_id_file']:
                self.authObj.setUUIDFile(config['client_id_file'])
        except Exception as e:
            self.logging.error(str(e))
            raise e

        try:
            if config['private_key']:
                self.authObj.setPrivateKey(config['private_key'])
        except Exception as e:
            self.logging.error(str(e))
            raise e

        try:
            if config['public_key']:
                self.authObj.setPublicKey(config['public_key'])
        except Exception as e:
            self.logging.error(str(e))
            raise e

    def get_task(self):
        try:
            return self.incoming_queue.get(False, 1)
        except:
            return None

    def set_task(self, task):
        return self.incoming_queue.put(task)

    def register_apps(self, appList):
        response = {
            "id": str(uuid4()),
            "cmd": "nodes",
            "route": "registerApps",
            "data": appList
        }
        self.send_response(json.dumps(response))

    def register_hybrid_tools(self, tool_catalog):
        response = {
            "id": str(uuid4()),
            "cmd": "nodes",
            "route": "registerHybridTools",
            "data": tool_catalog,
        }
        self.send_response(json.dumps(response))

    def register_profile(self, profile):
        response = {
            "id": str(uuid4()),
            "cmd": "nodes",
            "route": "registerProfile",
            "data": profile,
        }
        self.send_response(json.dumps(response))

    def subscribe_services(self, group_id=None):
        """Subscribe to service-only channel for hybrid_catalog messages.

        This must be called after WebSocket connection to receive hybrid_catalog
        result/submit messages on the {group}-services Redis channel, which is
        only subscribed to by service sessions (not human SOC users).

        The server will subscribe this session to all groups in the authenticated
        session; group_id parameter is optional for logging purposes.
        """
        if not group_id:
            group_id = self.authObj.getParentGroup() if self.authObj else None
        response = {
            "id": str(uuid4()),
            "cmd": "hybrid_catalog",
            "route": "subscribeServices",
            "group_id": group_id,
            "data": {},
        }
        self.logging.info("[hybrid] Subscribing to service channels for group: %s", group_id)
        self.send_response(json.dumps(response))

    def build_node_profile(self, available_tasks=None):
        available_tasks = available_tasks or []
        hostname = socket.gethostname()
        interfaces = []
        primary_ip = None
        primary_netmask = None

        try:
            for if_name, addrs in psutil.net_if_addrs().items():
                for addr in addrs:
                    if getattr(addr, "family", None) != socket.AF_INET:
                        continue
                    if addr.address.startswith("127."):
                        continue
                    entry = {
                        "name": if_name,
                        "ip": addr.address,
                        "netmask": addr.netmask,
                    }
                    interfaces.append(entry)
                    if primary_ip is None:
                        primary_ip = addr.address
                        primary_netmask = addr.netmask
        except Exception:
            pass

        return {
            "profile_id": self.uuid or str(uuid4()),
            "hostname": hostname,
            "ip": primary_ip,
            "netmask": primary_netmask,
            "interfaces": interfaces,
            "available_tasks": available_tasks,
        }

    def get_playbooks(self):

        url = random.choice(
            self.config['server']) + "/api/playbook/" + self.authObj.getParentGroup()
        try:
            response = self.authObj.Session.get(
                url, verify=self.authObj.verify)
        except Exception as e:
            self.logging.error("Failed fetching playbooks at url: %s" % url)
            self.logging.exception(e)
            raise e

        if response.status_code != 200:
            raise Exception("Status response not valid (%d): %s" %
                            (response.status_code, response.content))

        return response.json()

    def login(self):

        try:
            access_token = self.config.get('access_token') or self.config.get('access_key')
            secret_key = self.config.get('secret_key') or self.config.get('secret')

            if not access_token or not secret_key:
                raise Exception(
                    "Service-account authentication requires access_token and secret_key in config.")

            self.authObj.loginApiUser(
                access_token,
                secret_key,
            )
            self.logging.info(
                "Authenticated using service-account token pair; websocket session established.")

            self.uuid = self.authObj.getUUID()

            return True

        except Exception as e:
            self.logging.error(str(e))
            raise e

    def run(self, **args):
        self.run_loop(self.set_task)

    def run_loop(self, callback):
        # websocket
        if not self.authObj.Session:
            raise Exception("No valid authenticated session found, unable to continue.")

        self.isRunning = True

        def on_message(ws, message):
            try:
                data = json.loads(message)
            except Exception as e:
                print("Failed to parse json: %s" % message)
                return

            callback(data)

        def on_error(ws, error):
            self.logging.error("Websocket error detected: %s" % str(error))

        def on_close(ws, close_status_code=None, close_msg=None):
            self.logging.error("Websocket has closed. status=%s msg=%s" %
                               (close_status_code, close_msg))

        def on_open(ws):
            self.logging.info("WebSocket connected")
            # Subscribe to service-only channel for hybrid_catalog messages
            # This must be called after WebSocket connection to receive hybrid results
            try:
                self.subscribe_services()
            except Exception as e:
                self.logging.warning("[hybrid] Service channel subscription on connect failed: %s" % e)

            def run(*args):
                while self.isRunning:
                    # send any messages in our queue
                    try:
                        msg = self.outgoing_queue.get(False, 1)
                    except:
                        continue

                    if not msg:
                        continue

                    if type(msg) != str:
                        self.send_response(json.dumps(msg))
                    else:
                        self.send_response(msg)

                if ws:
                    try:
                        ws.close()
                    except Exception:
                        pass
                self.logging.info(
                    "Websocket communication thread terminating.")
            if self._sender_thread and self._sender_thread.is_alive():
                return
            self._sender_thread = threading.Thread(
                target=run,
                name="soarapi-ws-sender",
                daemon=True,
            )
            self._sender_thread.start()

        if self.config['debug']:
            websocket.enableTrace(True)
        ws_targets = self.config.get('websocket_server') or self.config['server']
        ws_target = random.choice(ws_targets).strip()
        if ws_target.endswith("/"):
            ws_target = ws_target[:-1]

        if ws_target.startswith("ws://") or ws_target.startswith("wss://"):
            connStr = ws_target if ws_target.endswith("/websocket") else ("%s/websocket" % ws_target)
        else:
            connStr = ws_target if ws_target.endswith("/websocket") else ("%s/websocket" % ws_target)
            if connStr.startswith("https://"):
                connStr = "wss://" + connStr[len("https://"):]
            elif connStr.startswith("http://"):
                connStr = "ws://" + connStr[len("http://"):]
        cookies = self.authObj.Session.cookies.get_dict()
        cookies_str = ""
        for k in cookies:
            cookies_str += "%s=%s; " % (k, cookies[k])
        self.logging.debug("Websocket connection string: %s (%s)" %
                           (connStr, cookies_str))

        websocket.enableTrace(False)
        self.ws = websocket.WebSocketApp(connStr,
                                         on_message=on_message,
                                         on_error=on_error,
                                         on_close=on_close,
                                         cookie=cookies_str)
        self.ws.on_open = on_open

        while self.isRunning:  # enables reconnects
            try:
                if self.config['insecure']:
                    self.ws.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})
                else:
                    self.ws.run_forever()
            except Exception as e:
                self.logging.error(
                    'soarAPI: websocket error occurred, reconnecting: %s' % e)
                time.sleep(1)

    def send_response(self, response):
        if self.ws:
            while self.isRunning:
                try:
                    if type(response) == str:
                        self.ws.send(response)
                    else:
                        if not 'id' in response:
                            response['id'] = str(uuid4())
                        self.ws.send(json.dumps(response))
                    # Log successful send for hybrid_task results (diagnostic)
                    if isinstance(response, dict) and response.get('cmd') == 'hybrid_task' and response.get('route') == 'result':
                        self.logging.info(
                            "[hybrid-result-sent] id=%s action=%s status=%s",
                            response.get('id'),
                            (response.get('data') or {}).get('action', 'unknown'),
                            (response.get('data') or {}).get('status', 'unknown')
                        )
                    break
                except Exception as e:
                    self.logging.warning("[hybrid-result-send-error] Exception during send: %s", e)
                    time.sleep(1)
        else:
            self.outgoing_queue.put(response)
        return True

    def set_pause(self):
        """Tells SOAR API that we are at capacity"""

        # not implement nice to have
        pass

    def set_resume(self):
        """Tells SOAR API that we can accept work again"""

        # not implement nice to have
        pass

    def close(self):
        self.authObj.finish()

    def shutdown(self):
        print("soar API shutdown...")
        self.isRunning = False
        if self.ws:
            self.ws.close()
        if self._sender_thread and self._sender_thread.is_alive():
            self._sender_thread.join(timeout=5)




import logging
import os
import sys
from time import time

import distro
import psutil

from hawkSoar.version import __VERSION__
from hawkSoar.common.logger import get_logger


class WorkloadStatistic:

    def __init__(self, config=None, api=None, component='ingest'):
        self.config = config
        self.component = component
        self.api = api
        self.platform_information = self._get_platform_information()
        self.logging = get_logger()
        self.logging.debug('{}: Online'.format(self.component))

    def shutdown(self):
        self.logging.debug(
            '{}: shutting down -> set offline message'.format(self.component))
        # self.api send_statistic
        # self.api.update_statistic(self.component, {'status': 'offline', 'last_update': time()})

    def update(self, ingest_workload=None, contain_workload=None, match_workload=None, decision_workload=None, investigate_workload=None, notify_workload=None, record_workload=None, recover_workload=None, utility_workload=None):
        stats = {
            'name': self.component,
            'status': 'online',
            'last_update': time(),
            'system': self._get_system_information(),
            'platform': self.platform_information,
        }

        if ingest_workload:
            stats['ingest'] = ingest_workload
        if contain_workload:
            stats['contain'] = contain_workload
        if match_workload:
            stats['match'] = match_workload
        if decision_workload:
            stats['decision'] = decision_workload
        if investigate_workload:
            stats['investigate'] = investigate_workload
        if notify_workload:
            stats['notify'] = notify_workload
        if record_workload:
            stats['record'] = record_workload
        if investigate_workload:
            stats['recover'] = recover_workload
        if utility_workload:
            stats['utility'] = utility_workload

        # self.api.update_statistic(self.component, stats)

    def _get_system_information(self):
        memory_usage = psutil.virtual_memory()
        try:
            disk_usage = psutil.disk_usage(
                self.config['data_storage']['firmware_file_storage_directory'])
        except Exception:
            disk_usage = psutil.disk_usage('/')
        try:
            cpu_percentage = psutil.cpu_percent()
        except Exception:
            cpu_percentage = 'unknown'

        result = {
            'cpu_cores': psutil.cpu_count(logical=False),
            'virtual_cpu_cores': psutil.cpu_count(),
            'cpu_percentage': cpu_percentage,
            'load_average': ', '.join(str(x) for x in os.getloadavg()),
            'memory_total': memory_usage.total,
            'memory_used': memory_usage.used,
            'memory_percent': memory_usage.percent,
            'disk_total': disk_usage.total,
            'disk_used': disk_usage.used,
            'disk_percent': disk_usage.percent
        }
        return result

    @staticmethod
    def _get_platform_information():
        operating_system = ' '.join(distro.linux_distribution()[0:2])
        python_version = '.'.join(str(x) for x in sys.version_info[0:3])
        version = __VERSION__
        return {
            'os': operating_system,
            'python': python_version,
            'version': version
        }



"""Remote deployment of hawk-soard to Windows target hosts via WinRM/PowerShell.

Provides script-building helpers called by HybridTaskExecutor._deploy_hawk_soard().
The deploying hawk-soard node uses PowerShell Invoke-Command (WinRM) to push
hawk-soard onto the destination host, configure it, and start it as a Windows service.

Two deployment strategies are supported:

  Push (preferred)
    When a local NSIS installer is found on disk (see ``find_local_installer``),
    it is copied to the remote host via a PSSession + ``Copy-Item -ToSession``
    and then run silently.  Credentials are forwarded to the NSIS installer
    via ``/ACCESS_TOKEN=`` and ``/SECRET_KEY=`` on the command line AND the
    config is written again via PowerShell afterwards to guarantee the correct
    values even on re-deployment (the NSIS installer skips config-writing when
    a previous config file already exists).

  Download
    When no local installer is available the remote host downloads the package
    from ``source_url``.  ``.exe`` packages receive the same NSIS CLI treatment.
    ``.zip`` packages fall back to manual config-write and service registration.

In both strategies the config is always (re-)written via PowerShell and the
service is always restarted so the new credentials take effect immediately.

Config flag required in hawk-soard.cfg to enable this action:
    enable_remote_deploy = true

Optional allowlist (CSV of hostnames/IPs):
    remote_deploy_allowed_targets = host1,host2,...
"""

import base64
import logging
import os

logger = logging.getLogger(__name__)

HAWK_DOWNLOAD_BASE = "https://downloads.hawkdefense.com/agent"
DEFAULT_SERVICE_NAME = "HawkSoard"
DEFAULT_INSTALL_DIR = "C:\\Program Files\\HAWK\\SOAR"
LOCAL_INSTALLER_FILENAME = "hawk-soard-setup.exe"
_SERVICE_DISPLAY_NAME = "Hawk SOAR Agent"
_SERVICE_DESCRIPTION = "Hawk Security Orchestration, Automation and Response daemon"


# ---------------------------------------------------------------------------
# Public helpers
# ---------------------------------------------------------------------------

def find_local_installer(config=None):
    """Return the path of a locally cached NSIS installer, or ``None``.

    Checks (in order):

    1. Directory containing the running binary.  In a PyInstaller-frozen exe
       ``sys.executable`` is the actual exe path; in development ``sys.argv[0]``
       is used.  When hawk-soard is installed via NSIS the installer copies
       itself there (``hawk-soard-setup.exe``) so it is always available for
       push-based remote deployment.
    2. Directory derived from the ``pid_file`` config key (the agent install
       directory as written by the NSIS installer).

    Args:
        config (dict, optional): The local agent's parsed config dict.

    Returns:
        str | None: Absolute path to the installer, or ``None`` when not found.
    """
    import sys

    candidates = []

    # Frozen PyInstaller exe: sys.executable is the real exe path.
    # Development: use sys.argv[0].
    if getattr(sys, "frozen", False):
        base = os.path.dirname(os.path.abspath(sys.executable))
    else:
        base = os.path.dirname(os.path.abspath(sys.argv[0]))
    candidates.append(os.path.join(base, LOCAL_INSTALLER_FILENAME))

    if config:
        pid_file = config.get("pid_file", "")
        if pid_file:
            install_dir = os.path.dirname(os.path.abspath(pid_file))
            candidate = os.path.join(install_dir, LOCAL_INSTALLER_FILENAME)
            if candidate not in candidates:
                candidates.append(candidate)

    for path in candidates:
        if os.path.isfile(path):
            logger.debug("find_local_installer: found %s" % path)
            return path

    return None


def build_config_content(hawk_server, access_token, secret_key, install_dir, insecure=False):
    """Build the hawk-soard.cfg INI file content for a newly deployed agent.

    Args:
        hawk_server:  HAWK platform URL (e.g. https://ir.hawk.io).
        access_token: Service-account access_token for the remote agent.
        secret_key:   Service-account secret_key for the remote agent.
        install_dir:  Install directory on the remote host.
        insecure:     If True, write ``insecure = true``.

    Returns:
        str: INI-formatted config file content.
    """
    idir = install_dir.rstrip("\\")
    return (
        "[SETTINGS]\n"
        "server = {server}\n"
        "access_token = {token}\n"
        "secret_key = {secret}\n"
        "insecure = {insecure}\n"
        "timeout = 30\n"
        "retry = 3\n"
        "pid_file = {idir}\\hawk-soard.pid\n"
        "private_key = {idir}\\hawk-soard.key\n"
        "public_key = {idir}\\hawk-soard.crt\n"
        "client_id_file = {idir}\\hawk-soard-id\n"
        "enforce_signed_hybrid_tasks = true\n"
        "enforce_group_id = true\n"
        "max_task_timeout = 300\n"
        "enable_command_execution = true\n"
        "auto_upgrade = true\n"
        "auto_upgrade_interval = 3600\n"
        "\n"
        "[MANAGER]\n"
        "poll = 1\n"
        "block = 1\n"
        "db_type = sqlite\n"
        "\n"
        "[PLUGINS]\n"
        "data_path = {idir}\\data\n"
    ).format(
        server=hawk_server,
        token=access_token,
        secret=secret_key,
        insecure="true" if insecure else "false",
        idir=idir,
    )


def build_deploy_script(
    target_host,
    user_ref,
    password,
    hawk_server,
    access_token,
    secret_key,
    install_dir=None,
    service_name=None,
    source_url=None,
    source_path=None,
    insecure=False,
    ps_quote_fn=None,
):
    """Build a PowerShell script that deploys hawk-soard to a remote Windows host.

    The returned string is a complete PowerShell command ready to pass to
    ``["powershell", "-NoProfile", "-NonInteractive", "-Command", script]``.

    Both strategies (push and download) unconditionally write the config file
    via PowerShell after the installer runs, then perform a service restart, so
    credentials are always current even when re-deploying to a host that already
    has hawk-soard installed.

    Args:
        target_host:  Remote hostname or IP address (must have WinRM enabled).
        user_ref:     WinRM credential user, e.g. ``"DOMAIN\\\\admin"``.
        password:     WinRM credential password.
        hawk_server:  HAWK platform URL written into the remote agent config.
        access_token: Service-account access_token for the remote agent.
        secret_key:   Service-account secret_key for the remote agent.
        install_dir:  Destination directory on the remote host.
                      Defaults to ``C:\\\\Program Files\\\\HAWK\\\\SOAR``.
        service_name: Windows service name.  Defaults to ``HawkSoard``.
        source_url:   URL the remote host downloads the hawk-soard package from.
                      Only used when ``source_path`` is not provided.
                      Defaults to the HAWK download server URL.
        source_path:  Local filesystem path to the NSIS installer on the
                      deploying host.  When provided the file is pushed to the
                      remote host via ``Copy-Item -ToSession`` (preferred).
        insecure:     Write ``insecure = true`` in the remote agent config.
        ps_quote_fn:  Optional callable ``(str) -> str`` for PowerShell string
                      escaping.  Uses the built-in single-quote escaper by default.

    Returns:
        str: Complete PowerShell command.
    """
    q = ps_quote_fn or _ps_quote
    idir = (install_dir or DEFAULT_INSTALL_DIR).rstrip("\\")
    svc = service_name or DEFAULT_SERVICE_NAME

    # Always build cfg_b64 for both strategies.  The NSIS installer has an
    # IfFileExists guard that skips config-writing when a prior config exists,
    # so we always write the config via PowerShell afterwards to ensure
    # credentials are current on re-deployment.
    cfg_text = build_config_content(hawk_server, access_token, secret_key, idir, insecure)
    cfg_b64 = base64.b64encode(cfg_text.encode("utf-8")).decode("ascii")

    if source_path and os.path.isfile(source_path):
        logger.info(
            "build_deploy_script: push strategy — local installer %s -> %s"
            % (source_path, target_host)
        )
        return _build_push_script(
            target_host, user_ref, password, access_token, secret_key,
            idir, svc, source_path, cfg_b64, q,
        )

    if not source_url:
        source_url = "%s/latest/%s" % (HAWK_DOWNLOAD_BASE.rstrip("/"), LOCAL_INSTALLER_FILENAME)

    logger.info(
        "build_deploy_script: download strategy — %s -> %s" % (source_url, target_host)
    )
    return _build_url_script(
        target_host, user_ref, password, access_token, secret_key,
        idir, svc, source_url, cfg_b64, q,
    )


def build_verify_script(target_host, user_ref, password, service_name=None, ps_quote_fn=None):
    """Build a lightweight PowerShell script that queries the remote service status.

    Returns:
        str: PowerShell command that emits a JSON service-status object.
    """
    q = ps_quote_fn or _ps_quote
    svc = service_name or DEFAULT_SERVICE_NAME

    remote_block = (
        "$svcName='%s';"
        "$s=Get-CimInstance Win32_Service -Filter (\"Name='\"+ $svcName +\"'\") -ErrorAction SilentlyContinue;"
        "if(-not $s){Write-Output '{\"status\":\"NotFound\",\"running\":false}';exit 0;};"
        "[PSCustomObject]@{"
        "service_name=$s.Name;status=$s.State;pid=$s.ProcessId;"
        "start_mode=$s.StartMode;running=($s.State -eq 'Running')"
        "}|ConvertTo-Json -Depth 2;"
    ) % q(svc)

    return (
        "$credSec=ConvertTo-SecureString '%s' -AsPlainText -Force;"
        "$cred=New-Object System.Management.Automation.PSCredential('%s',$credSec);"
        "Invoke-Command -ComputerName '%s' -Credential $cred -ScriptBlock {%s};"
    ) % (
        q(password),
        q(user_ref),
        q(target_host),
        remote_block,
    )


# ---------------------------------------------------------------------------
# Internal script builders
# ---------------------------------------------------------------------------

def _build_push_script(target_host, user_ref, password, access_token, secret_key,
                       idir, svc, source_path, cfg_b64, q):
    """PSSession + Copy-Item push strategy.

    1. Opens a PSSession to the remote host.
    2. Creates the install and data directories on the remote host.
    3. Copies the local NSIS installer to the remote host via Copy-Item -ToSession.
    4. Runs the installer using Start-Process so the /D= path-with-spaces
       argument is passed verbatim to NSIS (avoids PowerShell re-parsing it).
    5. Unconditionally writes the config via a base64-decoded blob so that
       credentials are current even when re-deploying over an existing install.
    6. Stops then starts the service so the new config takes effect.

    Credentials travel via -ArgumentList over the encrypted WinRM channel and
    are never embedded as string literals in the remote ScriptBlock.
    """
    # ScriptBlock 1: create directories before the file copy.
    setup_block = (
        "param($d);"
        "New-Item -ItemType Directory -Path $d -Force|Out-Null;"
        "New-Item -ItemType Directory -Path ($d+'\\data') -Force|Out-Null"
    )

    # ScriptBlock 2: run installer, write config, restart service.
    # Parameters: idir, svcName, tok (access_token), secKey (secret_key), cfgB64.
    # Note: $secKey avoids any name collision with the outer $credSec variable.
    install_block = (
        "param($idir,$svcName,$tok,$secKey,$cfgB64);"
        "$tmp=$idir+'\\hawk-soard-setup.exe';"
        "try{"
        # Use Start-Process so PowerShell does not re-tokenise the argument
        # string.  NSIS requires /D= to be the last argument and its path must
        # not be quoted — Start-Process -ArgumentList passes the string verbatim
        # to CreateProcess, which is exactly what NSIS expects.
        "$nsisArgs='/S /ACCESS_TOKEN='+$tok+' /SECRET_KEY='+$secKey+' /D='+$idir;"
        "Start-Process -FilePath $tmp -ArgumentList $nsisArgs -Wait -WindowStyle Hidden -ErrorAction Stop;"
        "Start-Sleep -Seconds 5;"
        # Always overwrite config to guarantee correct credentials on re-deploy.
        "$cfgBytes=[Convert]::FromBase64String($cfgB64);"
        "$cfgText=[Text.Encoding]::UTF8.GetString($cfgBytes);"
        "Set-Content -Path ($idir+'\\hawk-soard.cfg') -Value $cfgText -Encoding UTF8 -Force;"
        # Stop then start so the service reads the freshly written config.
        "Stop-Service -Name $svcName -Force -ErrorAction SilentlyContinue;"
        "Start-Sleep -Seconds 2;"
        "Start-Service -Name $svcName -ErrorAction Stop;"
        "Start-Sleep -Seconds 3;"
        "$s=Get-CimInstance Win32_Service -Filter (\"Name='\"+ $svcName +\"'\") -ErrorAction SilentlyContinue;"
        "$st=if($s){$s.State}else{'NotFound'};"
        "$spid=if($s){$s.ProcessId}else{0};"
        "[PSCustomObject]@{"
        "deployed=$true;service_name=$svcName;status=$st;pid=$spid;"
        "install_dir=$idir;exe_exists=(Test-Path ($idir+'\\hawk-soard.exe'));config_written=$true"
        "}|ConvertTo-Json -Depth 2;"
        "}catch{"
        "[PSCustomObject]@{deployed=$false;error=$_.Exception.Message;service_name=$svcName;install_dir=$idir}"
        "|ConvertTo-Json -Depth 2;}"
    )

    remote_installer_path = idir + "\\" + LOCAL_INSTALLER_FILENAME

    # Outer script: session lifecycle with catch so failures always emit JSON.
    # $credSec is used for the local SecureString to avoid any name clash with
    # the $secKey parameter inside the remote ScriptBlock.
    return (
        "$credSec=ConvertTo-SecureString '%s' -AsPlainText -Force;"
        "$cred=New-Object System.Management.Automation.PSCredential('%s',$credSec);"
        "$sess=New-PSSession -ComputerName '%s' -Credential $cred -ErrorAction Stop;"
        "try{"
        "Invoke-Command -Session $sess -ScriptBlock {%s} -ArgumentList '%s' -ErrorAction Stop;"
        "Copy-Item -Path '%s' -Destination '%s' -ToSession $sess -Force -ErrorAction Stop;"
        "Invoke-Command -Session $sess -ScriptBlock {%s} -ArgumentList '%s','%s','%s','%s','%s';"
        "}catch{"
        "[PSCustomObject]@{deployed=$false;error=$_.Exception.Message;service_name='%s';install_dir='%s'}"
        "|ConvertTo-Json -Depth 2;"
        "}finally{"
        "if($sess){Remove-PSSession $sess -ErrorAction SilentlyContinue}}"
    ) % (
        q(password),
        q(user_ref),
        q(target_host),
        setup_block,
        q(idir),
        q(source_path),
        q(remote_installer_path),
        install_block,
        q(idir),
        q(svc),
        q(access_token),
        q(secret_key),
        q(cfg_b64),
        q(svc),
        q(idir),
    )


def _build_url_script(target_host, user_ref, password, access_token, secret_key,
                      idir, svc, source_url, cfg_b64, q):
    """Self-contained Invoke-Command download-and-install strategy.

    Downloads the package from source_url on the remote host.

    - ``.exe`` packages: run via Start-Process with NSIS CLI args.
    - ``.zip`` packages: extract then register the service manually.

    In both cases the config is unconditionally written via a base64-decoded
    PowerShell blob and the service is restarted so credentials are always
    current, including on re-deployment.
    """
    remote_block = (
        "$idir='%s';"
        "$svcName='%s';"
        "$accessToken='%s';"
        "$secretKey='%s';"
        "$exePath=$idir+'\\hawk-soard.exe';"
        "$cfgPath=$idir+'\\hawk-soard.cfg';"
        "$srcUrl='%s';"
        "$cfgB64='%s';"
        # Derive temp-file extension from the URL, not the fixed filename, so
        # the .exe/.zip branch detection further down is always correct.
        "$tmpExt=if($srcUrl -match '\\.zip([?#]|$)'){'zip'}else{'exe'};"
        "$tmp=$env:TEMP+'\\hawk-soard-dl.'+$tmpExt;"
        "try{"
        # 1. Ensure directories exist.
        "New-Item -ItemType Directory -Path $idir -Force|Out-Null;"
        "New-Item -ItemType Directory -Path ($idir+'\\data') -Force|Out-Null;"
        # 2. Download and install only when the binary is not already present.
        "if(-not(Test-Path $exePath)){"
        "[Net.ServicePointManager]::SecurityProtocol=[Net.SecurityProtocolType]::Tls12;"
        "Invoke-WebRequest -Uri $srcUrl -OutFile $tmp -UseBasicParsing -ErrorAction Stop;"
        "if($tmpExt -eq 'exe'){"
        # Use Start-Process so the /D= path-with-spaces reaches NSIS verbatim.
        "$nsisArgs='/S /ACCESS_TOKEN='+$accessToken+' /SECRET_KEY='+$secretKey+' /D='+$idir;"
        "Start-Process -FilePath $tmp -ArgumentList $nsisArgs -Wait -WindowStyle Hidden -ErrorAction Stop;"
        "Start-Sleep -Seconds 5;"
        "}"
        "elseif($tmpExt -eq 'zip'){"
        "Expand-Archive -Path $tmp -DestinationPath $idir -Force;"
        # Manual service registration for zip installs.
        "$binPath='\"'+$exePath+'\" -c \"'+$cfgPath+'\" -service';"
        "$existing=Get-Service -Name $svcName -ErrorAction SilentlyContinue;"
        "if($existing){"
        "Stop-Service -Name $svcName -Force -ErrorAction SilentlyContinue;"
        "sc.exe config $svcName binPath= $binPath start= auto|Out-Null;}"
        "else{"
        "New-Service -Name $svcName -BinaryPathName $binPath "
        "-DisplayName '%s' -Description '%s' -StartupType Automatic|Out-Null;}"
        "};"
        "Remove-Item $tmp -Force -ErrorAction SilentlyContinue;"
        "};"
        # 3. Always write config — guarantees credentials are current even when
        #    the binary was already installed and the NSIS guard skipped writing.
        "$cfgBytes=[Convert]::FromBase64String($cfgB64);"
        "$cfgText=[Text.Encoding]::UTF8.GetString($cfgBytes);"
        "Set-Content -Path $cfgPath -Value $cfgText -Encoding UTF8 -Force;"
        # 4. Restart service so the new config takes effect.
        "Stop-Service -Name $svcName -Force -ErrorAction SilentlyContinue;"
        "Start-Sleep -Seconds 2;"
        "Start-Service -Name $svcName -ErrorAction Stop;"
        "Start-Sleep -Seconds 3;"
        # 5. Report status.
        "$s=Get-CimInstance Win32_Service -Filter (\"Name='\"+ $svcName +\"'\") -ErrorAction SilentlyContinue;"
        "$st=if($s){$s.State}else{'NotFound'};"
        "$spid=if($s){$s.ProcessId}else{0};"
        "[PSCustomObject]@{"
        "deployed=$true;service_name=$svcName;status=$st;pid=$spid;"
        "install_dir=$idir;exe_exists=(Test-Path $exePath);config_written=$true"
        "}|ConvertTo-Json -Depth 2;"
        "}catch{"
        "[PSCustomObject]@{"
        "deployed=$false;error=$_.Exception.Message;service_name=$svcName;install_dir=$idir"
        "}|ConvertTo-Json -Depth 2;"
        "}"
    ) % (
        q(idir),
        q(svc),
        q(access_token),
        q(secret_key),
        q(source_url),
        q(cfg_b64),
        q(_SERVICE_DISPLAY_NAME),
        q(_SERVICE_DESCRIPTION),
    )

    return (
        "$credSec=ConvertTo-SecureString '%s' -AsPlainText -Force;"
        "$cred=New-Object System.Management.Automation.PSCredential('%s',$credSec);"
        "Invoke-Command -ComputerName '%s' -Credential $cred -ScriptBlock {%s};"
    ) % (
        q(password),
        q(user_ref),
        q(target_host),
        remote_block,
    )


# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

def _ps_quote(value):
    """Escape a value for safe embedding inside a PowerShell single-quoted string."""
    if value is None:
        return ""
    return str(value).replace("'", "''")



# marketplace manager code.

# Check to see if this local inistance needs to install new plugin from marketplace.

from lib.marketplace import marketplaceAPI as mpAPI


class marketplaceManager(object):

    def __init__(self):
        pass

    def _get_new_plugins(self):
        return mpAPI.get_new_plugins()

    def install_plugin(self, plugin):
        """ Downloads plugin, then installs the plugin."""
        # download plugin mpAPI.download_plugin(uuid=plugin.uuid)

        # register plugin self.notify_plugin_manager(plugin)

        # if success return True, else return False
        pass

    def notify_plugin_manager(self):
        """ Notify plugin manager of new plugin. """

        # Notify plugin manager
        pass

    def run(self):

        while self.init:
            # check if new plugins are ready to be installed
            new_plugins = self._get_new_plugins()
            if new_plugins:
                for plugin in new_plugins:
                    self.install_plugin(plugin)
